import { world, system } from '@minecraft/server';
import CONFIG from '../config.js';
import NON_COLLIDABLE_BLOCKS from '../utils/nonCollidableBlocks.js';
import { getCachedBlock, clearBlockCache } from '../utils/blockCache.js';

class RollHandler {
    constructor(stateManager) {
        this.stateManager = stateManager;
        this.knockbackIntervals = new Map();
        this.solidBlockCache = new Map();
        this.lastTick = -1;
        this.lastLiquidBlockType = new Map();
        this.fallDamageTimers = new Map();
        this.rollInputTimers = new Map();
        this.fallDamageStates = new Map();
        this.rollActiveStates = new Map();
        this.fitOneBlockStates = new Map();
        this.effectTimers = new Map();
    }

    preWarmRoll(player) {
        const rollState = this.validateState(player);
        if (!rollState) return;

        try {
            player.setProperty('move:is_rolling', true);
            system.runTimeout(() => {
                player.playAnimation("animation.humanoid.roll", {
                    stopExpression: "!query.property('move:is_rolling')"
                });
                player.setProperty('move:is_rolling', false);
            });
        } catch (error) {
            console.error(`Pre-warm failed for ${player.name}: ${error}`);
        }
    }
	
    clearBlockCache() {
        clearBlockCache();
        this.solidBlockCache.clear();
    }

    getSolidBlockBelow(player, currentTick) {
        const { x, y, z } = player.location;
        const cacheKey = `${player.name}:${Math.floor(x)},${Math.floor(y)},${Math.floor(z)}`;
        const cached = this.solidBlockCache.get(cacheKey);

        if (cached && currentTick - cached.timestamp < 1) {
            return cached.result;
        }

        const offsets = [-0.29, 0, 0.29];
        let minDistance = Infinity;
        let solidBlock = null;

        outerLoop: for (const dx of offsets) {
            const checkX = Math.floor(x + dx);
            for (const dz of offsets) {
                const checkZ = Math.floor(z + dz);

                for (let i = 0; i < CONFIG.ROLL.CHECK_DISTANCE; i++) {
                    const yCheck = Math.floor(y) - 1 - i;
                    const block = getCachedBlock(player.dimension, { x: checkX, y: yCheck, z: checkZ });

                    if (!block) {
                        break;
                    }
                    if (!NON_COLLIDABLE_BLOCKS.includes(block.typeId)) {
                        const distance = Math.floor(y) - (block.location.y + 1);
                        if (distance < minDistance) {
                            minDistance = distance;
                            solidBlock = block;
                        }
                        break;
                    }
                }

                if (minDistance <= 1) {
                    break outerLoop;
                }
            }
        }

        const result = { block: solidBlock, distance: minDistance };
        this.solidBlockCache.set(cacheKey, { result, timestamp: currentTick });
        return result;
    }

    validateState(player) {
        return this.stateManager.getState(player.name)?.roll ?? null;
    }

    triggerFallDamageEvent(player, state) {
        const currentState = this.fallDamageStates.get(player.name);
        const newState = state === "move:fall_damage_true";

        if (currentState !== newState) {
            player.triggerEvent(state);
            this.fallDamageStates.set(player.name, newState);
        }
    }

    triggerFitOneBlockEvent(player, state) {
        const currentState = this.fitOneBlockStates.get(player.name);
        const newState = state === "move:fit_one_block_true";

        if (currentState !== newState) {
            player.triggerEvent(state);
            this.fitOneBlockStates.set(player.name, newState);
        }
    }

    checkLiquidCollision(player, dimension, location) {
        const pos = { x: Math.floor(location.x), y: Math.floor(location.y), z: Math.floor(location.z) };
        const block = getCachedBlock(dimension, pos);
        if (block?.isLiquid) {
            this.lastLiquidBlockType.set(player.name, block.typeId);
            return true;
        } else {
            this.lastLiquidBlockType.delete(player.name);
            return false;
        }
    }

    handleRoll(player, currentTick) {
        const rollState = this.validateState(player);
        if (!rollState || currentTick === this.lastTick) {
            return;
        }
        this.lastTick = currentTick;

        const { dimension, location } = player;

        if (!player.isOnGround) {
            this.handleAirState(player, rollState, currentTick, dimension, location);
            const rollActive = rollState.rollInputReceived || rollState.isRolling;
            const lastRollActive = this.rollActiveStates.get(player.name) ?? false;
            if (rollActive && !lastRollActive) {
                this.triggerFallDamageEvent(player, "move:fall_damage_false");
                this.rollActiveStates.set(player.name, true);
            } else if (!rollActive && lastRollActive) {
                this.rollActiveStates.set(player.name, false);
            }
        } else if (rollState.rollInputReceived) {
            this.handleLanding(player, rollState, currentTick, dimension, location);
            this.rollActiveStates.set(player.name, false);
        } else if (!rollState.isRolling) {
            this.resetFallState(rollState);
        }

        this.checkRollInputTimeout(player, rollState, currentTick);

        if (currentTick % 100 === 0) {
            this.clearBlockCache();
        }
    }

handleAirState(player, rollState, currentTick, dimension, location) {
    if (!rollState.isInAir) {
        rollState.isInAir = true;
        rollState.fallStartY = player.location.y;
        rollState.rollInputReceived = false;
        rollState.wasGliding = player.isGliding;
        rollState.wasSliding = player.isSneaking && player.isSprinting; 
    }

    const fallHeight = rollState.fallStartY - player.location.y;
    rollState.validFallHeight = fallHeight >= CONFIG.ROLL.FALL_HEIGHT_REQUIRED;

    const { distance: distanceToSolid } = this.getSolidBlockBelow(player, currentTick);

    if (!rollState.isRolling) {
        if (player.isJumping && rollState.validFallHeight && 
            !rollState.rollInputReceived && 
            distanceToSolid >= CONFIG.ROLL.ROLL_TRIGGER_MIN && 
            distanceToSolid <= CONFIG.ROLL.ROLL_TRIGGER_MAX) {
            rollState.rollInputReceived = true; 
            rollState.fallStartY = player.location.y;
            rollState.rollInputTime = system.currentTick;
            if (this.fallDamageTimers.has(player.name)) {
                system.clearRun(this.fallDamageTimers.get(player.name));
                this.fallDamageTimers.delete(player.name);
            }
            this.triggerFallDamageEvent(player, "move:fall_damage_false");
        }
        else if ((player.isGliding || rollState.wasGliding) && player.isJumping && 
                 !rollState.rollInputReceived && 
                 distanceToSolid <= CONFIG.ROLL.ROLL_TRIGGER_MAX) {
            rollState.rollInputReceived = true; 
            rollState.fallStartY = player.location.y;
            rollState.rollInputTime = system.currentTick;
            if (this.fallDamageTimers.has(player.name)) {
                system.clearRun(this.fallDamageTimers.get(player.name));
                this.fallDamageTimers.delete(player.name);
            }
            this.triggerFallDamageEvent(player, "move:fall_damage_false");
        }
    }

    if (player.isGliding) {
        rollState.wasGliding = true;
    }
    if (player.isSneaking && player.isSprinting) {
        rollState.wasSliding = true;
    } else if (player.isOnGround) {
        rollState.wasSliding = false;
    }
}

handleLanding(player, rollState, currentTick, dimension, location) {
    if (this.checkLiquidCollision(player, dimension, location)) {
        this.triggerFallDamageEvent(player, "move:fall_damage_true");
        this.resetFallState(rollState);
        return;
    }

    const { block: solidBlockBelow } = this.getSolidBlockBelow(player, currentTick);
    if (!rollState.isRolling && rollState.rollInputReceived && solidBlockBelow && 
        !NON_COLLIDABLE_BLOCKS.includes(solidBlockBelow.typeId)) {
        this.startRoll(player, currentTick, dimension, location);
        rollState.rollInputReceived = false;
    }
}

resetFallState(rollState) {
    rollState.isInAir = false;
    rollState.validFallHeight = false;
    rollState.rollInputReceived = false;
    rollState.fallStartY = 0;
    rollState.rollInputTime = undefined;
    rollState.wasGliding = false;
    rollState.wasSliding = false; 
}

    startContinuousKnockback(player) {
        const rollState = this.validateState(player);
        if (!rollState) return;

        if (this.knockbackIntervals.has(player.name)) {
            system.clearRun(this.knockbackIntervals.get(player.name));
            this.knockbackIntervals.delete(player.name);
        }

        const power = CONFIG.ROLL.KNOCKBACK_POWER;
        const reduction = power * (CONFIG.ROLL.KNOCKBACK_POWER_REDUCTION / 100);
        const totalTicks = CONFIG.ROLL.DURATION;
        let tickCount = 0;

        const intervalId = system.runInterval(() => {
            if (++tickCount >= totalTicks || power - reduction * tickCount <= 0 || !rollState.isRolling) {
                system.clearRun(intervalId);
                this.knockbackIntervals.delete(player.name);
                this.triggerFitOneBlockEvent(player, "move:fit_one_block_false");
                return;
            }

            const { y: yaw } = player.getRotation();
            const yawRad = (yaw + 90) * Math.PI / 180;
            const pos = player.location;
            const adjustedPower = power * (1 - tickCount / totalTicks);

            const offsets = [-0.29, 0, 0.29];
            let isOneBlockGap = false;

            outerLoop: for (const dx of offsets) {
                for (const dz of offsets) {
                    const checkX = Math.floor(pos.x + Math.cos(yawRad) + dx);
                    const checkZ = Math.floor(pos.z + Math.sin(yawRad) + dz);
                    const bottomBlock = getCachedBlock(player.dimension, { x: checkX, y: Math.floor(pos.y), z: checkZ });
                    const topBlock = getCachedBlock(player.dimension, { x: checkX, y: Math.floor(pos.y) + 1, z: checkZ });

                    const isBottomPassable = !bottomBlock || bottomBlock.typeId === "minecraft:air" || NON_COLLIDABLE_BLOCKS.includes(bottomBlock?.typeId);
                    const isTopSolid = topBlock && topBlock.typeId !== "minecraft:air" && !NON_COLLIDABLE_BLOCKS.includes(topBlock.typeId);

                    if (isTopSolid && isBottomPassable) {
                        isOneBlockGap = true;
                        break outerLoop;
                    }
                }
            }

            if (isOneBlockGap) {
                this.triggerFitOneBlockEvent(player, "move:fit_one_block_true");
            } else {
                this.triggerFitOneBlockEvent(player, "move:fit_one_block_false");
            }

            player.applyKnockback(
                Math.cos(yawRad) * adjustedPower,
                Math.sin(yawRad) * adjustedPower,
                adjustedPower,
                -0.05
            );
        }, 1);

        this.knockbackIntervals.set(player.name, intervalId);
    }

    checkRollInputTimeout(player, rollState, currentTick) {
        if (rollState.rollInputReceived && !rollState.isRolling && rollState.rollInputTime !== undefined) {
            const ticksSinceInput = system.currentTick - rollState.rollInputTime;
            if (ticksSinceInput >= 20) {
                this.resetFallState(rollState);
                if (this.rollInputTimers.has(player.name)) {
                    system.clearRun(this.rollInputTimers.get(player.name));
                    this.rollInputTimers.delete(player.name);
                }
            } else if (!this.rollInputTimers.has(player.name)) {
                const timerId = system.runTimeout(() => {
                    if (rollState.rollInputReceived && !rollState.isRolling) {
                        this.resetFallState(rollState);
                    }
                    this.rollInputTimers.delete(player.name);
                }, 20 - ticksSinceInput);
                this.rollInputTimers.set(player.name, timerId);
            }
        }
    }

    startRoll(player, currentTick, dimension, location) {
        const rollState = this.validateState(player);
        if (!rollState) return;

        if (this.checkLiquidCollision(player, dimension, location)) return;

        this.resetFallState(rollState);

        rollState.isRolling = true;
        rollState.rollStartTime = currentTick;
        rollState.lastUpdateTime = currentTick;

        if (this.rollInputTimers.has(player.name)) {
            system.clearRun(this.rollInputTimers.get(player.name));
            this.rollInputTimers.delete(player.name);
        }

        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }
        if (this.fallDamageTimers.has(player.name)) {
            system.clearRun(this.fallDamageTimers.get(player.name));
            this.fallDamageTimers.delete(player.name);
        }

        this.triggerFallDamageEvent(player, "move:fall_damage_false");
        this.triggerFitOneBlockEvent(player, "move:fit_one_block_false");
        player.setProperty('move:is_rolling', true);
        player.dimension.spawnParticle("move:drop", player.location);

        player.playSound(CONFIG.ROLL.SOUND_EFFECT.NAME, {
            volume: CONFIG.ROLL.SOUND_EFFECT.VOLUME,
            pitch: CONFIG.ROLL.SOUND_EFFECT.PITCH
        });

        const animationTimerId = system.runTimeout(() => {
            player.playAnimation("animation.humanoid.roll", {
                stopExpression: "!query.property('move:is_rolling')",
                blendOutTime: 1
            });
            this.effectTimers.delete(player.name);
        }, 1);
        this.effectTimers.set(player.name, animationTimerId);

        this.stateManager.setRollState(player.name, currentTick);
        this.startContinuousKnockback(player);

        const endTimerId = system.runTimeout(() => {
            this.endRoll(player, system.currentTick);
            this.fallDamageTimers.delete(player.name);
        }, CONFIG.ROLL.DURATION);
        this.fallDamageTimers.set(player.name, endTimerId);
    }

    endRoll(player, currentTick) {
        const rollState = this.validateState(player);
        if (!rollState) return;

        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }
        if (this.fallDamageTimers.has(player.name)) {
            system.clearRun(this.fallDamageTimers.get(player.name));
            this.fallDamageTimers.delete(player.name);
        }
        if (this.knockbackIntervals.has(player.name)) {
            system.clearRun(this.knockbackIntervals.get(player.name));
            this.knockbackIntervals.delete(player.name);
        }

        rollState.isRolling = false;
        rollState.lastRollTime = currentTick;

        player.setProperty('move:is_rolling', false);
        this.triggerFitOneBlockEvent(player, "move:fit_one_block_false");

        const timerId = system.runTimeout(() => {
            if (!rollState.isRolling && !rollState.rollInputReceived) {
                this.triggerFallDamageEvent(player, "move:fall_damage_true");
            }
            this.fallDamageTimers.delete(player.name);
        }, 10);
        this.fallDamageTimers.set(player.name, timerId);

        this.resetFallState(rollState);
        this.stateManager.clearRollState(player.name, currentTick);
    }
}

export default RollHandler;