import { MolangVariableMap, world, system } from '@minecraft/server';
import CONFIG from '../config.js';
import { BLOCK_SOUND_MAPPING, SOUND_CONFIGS, DEFAULT_SOUND } from '../utils/Data.js';
import NON_COLLIDABLE_BLOCKS from '../utils/nonCollidableBlocks.js';
import { getCachedBlock, clearBlockCache } from '../utils/blockCache.js';

class SlideHandler {
    constructor(stateManager) {
        this.stateManager = stateManager;
        this.solidBlockCache = new Map();
        this.lastTick = -1;
        this.effectTimers = new Map();
    }

getSolidBlockBelow(player, currentTick) {
    const { x, y, z } = player.location;
    const dimension = player.dimension;
    const cacheKey = `${player.name}:${Math.floor(x)},${Math.floor(y)},${Math.floor(z)}`;
    const cached = this.solidBlockCache.get(cacheKey);

    if (cached && currentTick - cached.timestamp < 1) {
        return cached.result;
    }

    const offsets = [-0.29, 0, 0.29];
    const feetY = Math.floor(y); 
    let solidBlock = null;
    let minDistance = Infinity;

    for (const dx of offsets) {
        const checkX = Math.floor(x + dx);
        for (const dz of offsets) {
            const checkZ = Math.floor(z + dz);
            const block = getCachedBlock(dimension, { x: checkX, y: feetY, z: checkZ });

            if (block && !NON_COLLIDABLE_BLOCKS.includes(block.typeId)) {
                const distance = y - (block.location.y + 1);
                if (distance < minDistance) {
                    minDistance = distance;
                    solidBlock = block;
                }
            }
        }
    }

    if (!solidBlock) {
        const yCheck = feetY - 1;
        for (const dx of offsets) {
            const checkX = Math.floor(x + dx);
            for (const dz of offsets) {
                const checkZ = Math.floor(z + dz);
                const block = getCachedBlock(dimension, { x: checkX, y: yCheck, z: checkZ });

                if (block && !NON_COLLIDABLE_BLOCKS.includes(block.typeId)) {
                    const distance = y - (block.location.y + 1);
                    if (distance < minDistance) {
                        minDistance = distance;
                        solidBlock = block;
                    }
                }
            }
        }
    }

    const result = {
        block: solidBlock,
        distance: solidBlock ? minDistance : 2
    };
    this.solidBlockCache.set(cacheKey, { result, timestamp: currentTick });
    return result;
}

    validateState(player) {
        const state = this.stateManager.getState(player.name);
        return state?.slide ?? null;
    }

    clearCaches() {
        clearBlockCache();
        this.solidBlockCache.clear();
    }

    checkOneBlockGap(player, currentTick) {
        const { location, dimension } = player;
        const { y: yaw } = player.getRotation();
        const yawRad = (yaw + 90) * Math.PI / 180;

        const offsets = [-0.29, 0, 0.29];
        let isOneBlockGap = false;

        outerLoop: for (const dx of offsets) {
            for (const dz of offsets) {
                const checkX = Math.floor(location.x + Math.cos(yawRad) + dx);
                const checkZ = Math.floor(location.z + Math.sin(yawRad) + dz);
                const bottomBlock = getCachedBlock(dimension, { x: checkX, y: Math.floor(location.y), z: checkZ });
                const topBlock = getCachedBlock(dimension, { x: checkX, y: Math.floor(location.y) + 1, z: checkZ });

                const isBottomPassable = !bottomBlock || bottomBlock.typeId === "minecraft:air" || NON_COLLIDABLE_BLOCKS.includes(bottomBlock?.typeId);
                const isTopSolid = topBlock && topBlock.typeId !== "minecraft:air" && !NON_COLLIDABLE_BLOCKS.includes(topBlock.typeId);

                if (isTopSolid && isBottomPassable) {
                    isOneBlockGap = true;
                    break outerLoop;
                }
            }
        }

        return isOneBlockGap;
    }

    isInInvalidState(player, dimension, location) {
        const pos = { x: Math.floor(location.x), y: Math.floor(location.y), z: Math.floor(location.z) };
        const block = getCachedBlock(dimension, pos);
        return block?.isLiquid;
    }

    handleSlide(player, currentTick) {
        if (currentTick === this.lastTick) return;
        this.lastTick = currentTick;

        const slide = this.validateState(player);
        if (!slide) return;

        const state = this.stateManager.getState(player.name);
        const { dimension, location } = player;

        if (slide.lastSlideTime && currentTick - slide.lastSlideTime < CONFIG.SLIDE.COOLDOWN) return;

        if (player.isSprinting) {
            slide.wasSprintingFirst = true;
            slide.lastSprintTime = currentTick;
        }

        const sprintBuffer = CONFIG.SLIDE.SPRINT_BUFFER;
        const wasRecentlySprinting = currentTick - slide.lastSprintTime < sprintBuffer;
        if (!wasRecentlySprinting && !player.isSneaking) slide.wasSprintingFirst = false;

        if (wasRecentlySprinting && player.isSneaking && slide.wasSprintingFirst && !slide.isSliding && player.isOnGround) {
            this.startSlide(player, currentTick, dimension, location);
        }

        if (slide.isSliding) {
            this.updateSlideMovement(player, currentTick, state, dimension, location);
        }

        if (currentTick % 100 === 0) this.clearCaches();
    }

    startSlide(player, currentTick, dimension, location) {
        const state = this.stateManager.getState(player.name);
        if (!state || state.roll?.isRolling) return;

        if (this.isInInvalidState(player, dimension, location)) return;
        const { block: solidBlock } = this.getSolidBlockBelow(player, currentTick);
        if (!solidBlock) return;

        state.slide.isSliding = true;
        state.slide.startTime = currentTick;
        player.triggerEvent("move:fit_one_block_true");
        player.setProperty('move:is_sliding', true);

        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }

        const timerId = system.runTimeout(() => {
            player.playAnimation("animation.humanoid.slide", {
                stopExpression: "!query.is_sneaking || !query.property('move:is_sliding')",
                blendOutTime: 0.3,
            });
            this.effectTimers.delete(player.name);
        }, 1);
        this.effectTimers.set(player.name, timerId);

        this.stateManager.setSlideState(player.name, {
            startTime: currentTick,
            slideDuration: CONFIG.SLIDE.DURATION,
            currentKnockback: CONFIG.SLIDE.KNOCKBACK_POWER,
            lastUpdateTime: currentTick,
        });
    }

    updateSlideMovement(player, currentTick, state, dimension, location) {
        const slideInfo = state.slide.slideInfo;
        if (!slideInfo) return;

        if (this.isInInvalidState(player, dimension, location)) {
            this.endSlide(player, currentTick);
            return;
        }

        if (currentTick - slideInfo.startTime > CONFIG.SLIDE.DURATION || !player.isSneaking) {
            this.endSlide(player, currentTick);
            return;
        }

        const timeSinceLastUpdate = (currentTick - slideInfo.lastUpdateTime) / 20;
        slideInfo.currentKnockback *= Math.pow(1 - (CONFIG.SLIDE.KNOCKBACK_POWER_REDUCTION / 100), timeSinceLastUpdate);
        slideInfo.lastUpdateTime = currentTick;

        const { y: yaw } = player.getRotation();
        const yawRad = (yaw + 90) * Math.PI / 180;

        // Dynamic gravity based on distance to solid block below (1 block scan)
        const { distance } = this.getSolidBlockBelow(player, currentTick);
        let elevation;
        if (distance <= 0.1) {
            // On or very close to ground (e.g., grass), slight upward force for smooth sliding
            elevation = CONFIG.SLIDE.ELEVATION.GROUND; // 0.075
        } else {
            // Above ground (e.g., air, water, or edge), apply downward force proportional to distance
            elevation = CONFIG.SLIDE.ELEVATION.AIR * Math.min(distance, 1); // Scales up to -0.13
        }

        const knockbackX = Math.cos(yawRad) * slideInfo.currentKnockback;
        const knockbackZ = Math.sin(yawRad) * slideInfo.currentKnockback;
        slideInfo.lastKnockbackTime = slideInfo.lastKnockbackTime || currentTick;
        player.applyKnockback(knockbackX, knockbackZ, slideInfo.currentKnockback, elevation);
        slideInfo.lastKnockbackTime = currentTick;

        const blockBelow = getCachedBlock(dimension, { x: Math.floor(location.x), y: Math.floor(location.y) - 1, z: Math.floor(location.z) });
        const blockSound = BLOCK_SOUND_MAPPING[blockBelow?.typeId] || DEFAULT_SOUND.sound;
        const soundConfig = SOUND_CONFIGS[blockSound] || DEFAULT_SOUND;
        dimension.playSound(blockSound, location, { pitch: soundConfig.pitch, volume: soundConfig.volume });

        const isOneBlockGap = this.checkOneBlockGap(player, currentTick);
        if (isOneBlockGap) {
            player.triggerEvent("move:fit_one_block_true");
        } else {
            player.triggerEvent("move:fit_one_block_false");
        }

        if (CONFIG.SLIDE.FORCE_CRAWLING) this.forceCrawling(player);
        this.spawnParticle(player);
    }

    spawnParticle(player) {
        const playerVelocity = player.getVelocity();
        const velocity = Math.sqrt(playerVelocity.x ** 2 + playerVelocity.z ** 2);

        if (velocity > 0.15) {
            const { location, dimension } = player;
            const { y: yaw } = player.getRotation();
            const rotationRad = (yaw + 90) * Math.PI / 180;
            const dirX = Math.cos(rotationRad);
            const dirZ = Math.sin(rotationRad);

            const molang = new MolangVariableMap();
            molang.setFloat("variable.direction_x", dirX);
            molang.setFloat("variable.direction_y", 0);
            molang.setFloat("variable.direction_z", dirZ);

            const spawnPos = {
                x: location.x - (playerVelocity.x * 0.3),
                y: location.y + 0.1,
                z: location.z - (playerVelocity.z * 0.3),
            };

            dimension.spawnParticle("move:trail", spawnPos, molang);
        }
    }

    forceCrawling(player) {
        const { location, dimension } = player;
        const { x: dirX, z: dirZ } = player.getViewDirection();
        const baseX = Math.floor(location.x);
        const baseY = Math.round(location.y) + 1;
        const baseZ = Math.floor(location.z);
        const barrierPositions = [];

        const range = CONFIG.SLIDE.BARRIER_RANGE || 2;
        for (let offsetX = -range; offsetX <= range + 1; offsetX++) {
            for (let offsetZ = -range; offsetZ <= range + 1; offsetZ++) {
                const blockPos = {
                    x: baseX + offsetX,
                    y: baseY,
                    z: baseZ + (offsetZ * dirZ) + (offsetZ * dirX),
                };
                const block = getCachedBlock(dimension, blockPos);
                if (block?.typeId === "minecraft:air") {
                    block.setType("minecraft:barrier");
                    barrierPositions.push(blockPos);
                }
            }
        }

        const timeout = CONFIG.SLIDE.FORCE_CRAWLING_TIMEOUT || 8;
        system.runTimeout(() => {
            barrierPositions.forEach(pos => {
                const block = getCachedBlock(dimension, pos);
                if (block?.typeId === "minecraft:barrier") block.setType("minecraft:air");
            });
        }, timeout);
    }

    endSlide(player, currentTick) {
        const state = this.stateManager.getState(player.name);
        if (!state) return;

        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }

        state.slide.isSliding = false;
        state.slide.lastSlideTime = currentTick;
        player.triggerEvent("move:fit_one_block_false");
        player.setProperty('move:is_sliding', false);
        this.stateManager.clearSlideState(player.name, currentTick);
    }
}

export default SlideHandler;