import { world, system } from '@minecraft/server';
import CONFIG from '../config.js';
import NON_COLLIDABLE_BLOCKS from '../utils/nonCollidableBlocks.js';
import { getCachedBlock, clearBlockCache } from '../utils/blockCache.js';

const isSolidBlock = (block) => block?.permutation && !NON_COLLIDABLE_BLOCKS.includes(block.typeId);

class VaultHandler {
    constructor(stateManager) {
        this.stateManager = stateManager;
        this.fallDamageTimers = new Map();
        this.lastTick = -1;
        this.effectTimers = new Map();
    }

    validateState(player) {
        const state = this.stateManager.getState(player.name);
        return state?.vault ?? null;
    }

    clearCaches() {
        clearBlockCache();
    }

    isInInvalidState(player, dimension, location) {
        const pos = { x: Math.floor(location.x), y: Math.floor(location.y), z: Math.floor(location.z) };
        const block = getCachedBlock(dimension, pos);
        return block?.isLiquid;
    }

    handleVault(player, currentTick) {
        if (currentTick === this.lastTick) return;
        this.lastTick = currentTick;

        const vaultState = this.validateState(player);
        if (!vaultState) return;

        if (vaultState.isVaulting) {
            if (currentTick - vaultState.vaultStartTime > CONFIG.VAULT.DURATION) {
                this.endVault(player, currentTick);
            }
            return;
        }

        if (vaultState.lastVaultTime && (currentTick - vaultState.lastVaultTime) < CONFIG.VAULT.COOLDOWN) {
            vaultState.cooldownNotified = true;
            return;
        }
        vaultState.cooldownNotified = false;

        if (!player.isSprinting || !player.isJumping) return;

        const state = this.stateManager.getState(player.name);
        const dimension = world.getDimension(player.dimension.id);
        if (state.roll?.isRolling || state.slide?.isSliding || state.wallrun?.isWallRunning) {
            vaultState.preventionNotified = true;
            return;
        }
        vaultState.preventionNotified = false;

        this.startVault(player, currentTick, state, dimension);

        if (currentTick % 100 === 0) this.clearCaches();
    }

    startVault(player, currentTick, state, dimension) {
        const vaultState = this.validateState(player);
        if (!vaultState) return;

        const { x, y, z } = player.location;
        const viewDir = player.getViewDirection();

        if (this.isInInvalidState(player, dimension, { x, y, z })) return;

        let foundSolid = false, tooHigh = false, tooLow = true;
        const minVaultHeight = y + CONFIG.VAULT.MIN_HEIGHT;
        const maxVaultHeight = y + CONFIG.VAULT.MAX_HEIGHT;
        const scanLimit = y + 3;

        for (let scanY = y + 0.5; scanY <= scanLimit; scanY += 0.5) {
            const block = getCachedBlock(dimension, {
                x: Math.floor(x + viewDir.x),
                y: Math.floor(scanY),
                z: Math.floor(z + viewDir.z),
            });
            if (isSolidBlock(block)) {
                foundSolid = true;
                if (scanY < minVaultHeight) {
                    tooLow = true;
                } else if (scanY > maxVaultHeight) {
                    tooHigh = true;
                    break;
                } else {
                    tooLow = false;
                }
            }
        }

        if (!foundSolid || tooLow || tooHigh) {
            vaultState.tooHighNotified = tooHigh;
            return;
        }
        vaultState.tooHighNotified = false;

        Object.assign(vaultState, { 
            isVaulting: true, 
            vaultStartTime: currentTick, 
            lastVaultTime: currentTick 
        });

        this.stateManager.setVaultState(player.name, null);

        if (state.roll) {
            this.resetFallState(state.roll);
            player.triggerEvent("move:fall_damage_false");
        }

        const boostStrength = CONFIG.VAULT.DETECTION_RADIUS;
        const upwardBoost = 0.52;
        const forwardVector = player.getViewDirection();

        player.setProperty("move:is_vaulting", true);
        player.playSound(CONFIG.VAULT.SOUND_EFFECT.NAME, {
            pitch: CONFIG.VAULT.SOUND_EFFECT.PITCH,
            volume: CONFIG.VAULT.SOUND_EFFECT.VOLUME,
        });

        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }
        const timerId = system.runTimeout(() => {
            player.playAnimation("animation.humanoid.vault", {
                stopExpression: "!query.property('move:is_vaulting')",
                blendOutTime: 1
            });
            this.effectTimers.delete(player.name);
        }, 1);
        this.effectTimers.set(player.name, timerId);

        player.applyKnockback(
            forwardVector.x * boostStrength,
            forwardVector.z * boostStrength,
            0.2,
            upwardBoost
        );
    }

    resetFallState(rollState) {
        Object.assign(rollState, {
            isInAir: false,
            validFallHeight: false,
            rollInputReceived: false,
            fallStartY: 0
        });
    }

    endVault(player, currentTick) {
        const state = this.stateManager.getState(player.name);
        if (!state) return;

        const vaultState = this.validateState(player);
        if (!vaultState) return;

        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }
        if (this.fallDamageTimers.has(player.name)) {
            system.clearRun(this.fallDamageTimers.get(player.name));
            this.fallDamageTimers.delete(player.name);
        }

        vaultState.isVaulting = false;
        player.setProperty("move:is_vaulting", false);
        this.stateManager.clearVaultState(player.name, currentTick);

        if (state.roll && !state.roll.isRolling && !state.roll.rollInputReceived) {
            const timerId = system.runTimeout(() => {
                if (!state.roll?.isRolling && !state.roll?.rollInputReceived) {
                    player.triggerEvent("move:fall_damage_true");
                }
                this.fallDamageTimers.delete(player.name);
            }, 10);
            this.fallDamageTimers.set(player.name, timerId);
        }
    }
}

export default VaultHandler;