import { world, system } from '@minecraft/server';
import CONFIG from '../config.js';
import { BLOCK_SOUND_MAPPING, SOUND_CONFIGS, DEFAULT_SOUND } from '../utils/Data.js';
import NON_COLLIDABLE_BLOCKS from '../utils/nonCollidableBlocks.js';
import { getCachedBlock, clearBlockCache } from '../utils/blockCache.js';

const isSolidBlock = (block) => block?.permutation && !NON_COLLIDABLE_BLOCKS.includes(block.typeId);

class WallRunHandler {
    constructor(stateManager) {
        this.stateManager = stateManager;
        this.fallDamageTimers = new Map();
        this.lastTick = -1;
        this.effectTimers = new Map();
    }

    validateState(player) {
        const state = this.stateManager.getState(player.name);
        return state?.wallrun ?? null;
    }

    clearCaches() {
        clearBlockCache();
    }

    getHeightAboveGround(player, dimension) {
        const { x, y, z } = player.location;
        const floorX = Math.floor(x);
        const floorZ = Math.floor(z);
        const startY = Math.floor(y - 0.5);

        for (let i = 0; i <= 10; i++) {
            const block = getCachedBlock(dimension, { x: floorX, y: startY - i, z: floorZ });
            if (isSolidBlock(block)) return y - (startY - i + 1);
        }
        return y - (startY - 10);
    }

    isInInvalidState(player, dimension, location) {
        const pos = { x: Math.floor(location.x), y: Math.floor(location.y), z: Math.floor(location.z) };
        const block = getCachedBlock(dimension, pos);
        return block?.isLiquid;
    }

    handleWallRun(player, currentTick) {
        if (currentTick === this.lastTick) return;
        this.lastTick = currentTick;

        const wallrun = this.validateState(player);
        if (!wallrun) return;

        if (wallrun.isWallRunning) {
            this.updateWallRun(player, currentTick);
            return;
        }

        if (wallrun.lastWallRunTime && (currentTick - wallrun.lastWallRunTime) < CONFIG.WALLRUN.COOLDOWN) return;

        const state = this.stateManager.getState(player.name);
        const { dimension, location } = player;
        const isBlocked = state.vault?.isVaulting || state.slide?.isSliding || state.roll?.isRolling;
        if (isBlocked || !player.isSprinting || player.isOnGround || player.isGliding) return;

        this.startWallRun(player, currentTick, state, dimension, location);

        if (currentTick % 100 === 0) this.clearCaches();
    }

    startWallRun(player, currentTick, state, dimension, location) {
        const wallrun = this.validateState(player);
        if (!wallrun) return;

        const heightAboveGround = this.getHeightAboveGround(player, dimension);
        if (heightAboveGround < CONFIG.WALLRUN.MIN_HEIGHT_ABOVE_GROUND) return;

        if (this.isInInvalidState(player, dimension, location)) return;

        const viewDir = player.getViewDirection();
        const yawRad = Math.atan2(viewDir.z, viewDir.x);
        const leftDirX = Math.cos(yawRad + Math.PI / 2);
        const leftDirZ = Math.sin(yawRad + Math.PI / 2);
        const rightDirX = Math.cos(yawRad - Math.PI / 2);
        const rightDirZ = Math.sin(yawRad - Math.PI / 2);

        const detectionRadius = CONFIG.WALLRUN.DETECTION_RADIUS;
        const checkPosLeft = { x: location.x + leftDirX * detectionRadius, y: location.y, z: location.z + leftDirZ * detectionRadius };
        const checkPosRight = { x: location.x + rightDirX * detectionRadius, y: location.y, z: location.z + rightDirZ * detectionRadius };
        const checkPosFront = { x: location.x + viewDir.x * detectionRadius, y: location.y, z: location.z + viewDir.z };

        const blockLeft = getCachedBlock(dimension, { x: Math.floor(checkPosLeft.x), y: Math.floor(checkPosLeft.y), z: Math.floor(checkPosLeft.z) });
        const blockRight = getCachedBlock(dimension, { x: Math.floor(checkPosRight.x), y: Math.floor(checkPosRight.y), z: Math.floor(checkPosRight.z) });
        const blockFront = getCachedBlock(dimension, { x: Math.floor(checkPosFront.x), y: Math.floor(checkPosFront.y), z: Math.floor(checkPosFront.z) });

        let side = null;
        if (isSolidBlock(blockLeft)) side = "left";
        else if (isSolidBlock(blockRight)) side = "right";

        if (!side || isSolidBlock(blockFront)) return;

        const knockbackPower = CONFIG.WALLRUN.KNOCKBACK_POWER;

        this.stateManager.setWallRunState(player.name, side, {
            startTime: currentTick,
            currentKnockback: knockbackPower,
            lastUpdateTime: currentTick,
            lastSoundTime: 0,
            lastParticleTime: 0,
            lastFootParticleTime: 0
        });

        player.setProperty("move:is_wallrunning", true);
        player.triggerEvent("move:fall_damage_false");

        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }
        const animation = side === "left" ? "animation.humanoid.wallrun.left" : "animation.humanoid.wallrun.right";
        const timerId = system.runTimeout(() => {
            player.playAnimation(animation, {
                stopExpression: "!query.property('move:is_wallrunning')",
                blendOutTime: 0.5
            });
            this.effectTimers.delete(player.name);
        }, 1);
        this.effectTimers.set(player.name, timerId);

        player.applyKnockback(viewDir.x * knockbackPower, viewDir.z * knockbackPower, knockbackPower, 0);
    }

    updateWallRun(player, currentTick) {
        const wallrun = this.validateState(player);
        if (!wallrun || !wallrun.wallRunInfo) return;
        const wallRunInfo = wallrun.wallRunInfo;

        const { location, dimension } = player;
        const detectionRadius = CONFIG.WALLRUN.DETECTION_RADIUS;

        const viewDir = player.getViewDirection();
        const blockFront = getCachedBlock(dimension, {
            x: Math.floor(location.x + viewDir.x * detectionRadius),
            y: Math.floor(location.y),
            z: Math.floor(location.z + viewDir.z)
        });
        if (isSolidBlock(blockFront)) {
            this.endWallRun(player, currentTick);
            return;
        }

        if (player.isOnGround || (currentTick - wallRunInfo.startTime) > CONFIG.WALLRUN.DURATION || !player.isSprinting || player.isGliding) {
            this.endWallRun(player, currentTick);
            return;
        }

        if (this.isInInvalidState(player, dimension, location)) {
            this.endWallRun(player, currentTick);
            return;
        }

        const yawRad = Math.atan2(viewDir.z, viewDir.x);
        const sideDirX = wallrun.side === "left" ? Math.cos(yawRad + Math.PI / 2) : Math.cos(yawRad - Math.PI / 2);
        const sideDirZ = wallrun.side === "left" ? Math.sin(yawRad + Math.PI / 2) : Math.sin(yawRad - Math.PI / 2);
        const blockSide = getCachedBlock(dimension, {
            x: Math.floor(location.x + sideDirX * detectionRadius),
            y: Math.floor(location.y),
            z: Math.floor(location.z + sideDirZ)
        });
        if (!isSolidBlock(blockSide)) {
            this.endWallRun(player, currentTick);
            return;
        }

        const timeSinceLastUpdate = (currentTick - wallRunInfo.lastUpdateTime) / 20;
        wallRunInfo.currentKnockback *= Math.pow(1 - (CONFIG.WALLRUN.KNOCKBACK_REDUCTION / 100), timeSinceLastUpdate);
        wallRunInfo.lastUpdateTime = currentTick;

        const forwardX = viewDir.x * wallRunInfo.currentKnockback;
        const forwardZ = viewDir.z * wallRunInfo.currentKnockback;
        const gravity = -0.03;
        const lateralDriftMagnitude = 0.2;
        const lateralDriftX = sideDirX * lateralDriftMagnitude;
        const lateralDriftZ = sideDirZ * lateralDriftMagnitude;
        player.applyKnockback(forwardX + lateralDriftX, forwardZ + lateralDriftZ, wallRunInfo.currentKnockback, gravity);

        const handOffsetY = 1.1;
        const handOffsetDistance = 0.3;
        const handOffsetX = sideDirX * handOffsetDistance;
        const handOffsetZ = sideDirZ * handOffsetDistance;
        const handParticlePos = { x: location.x + handOffsetX, y: location.y + handOffsetY, z: location.z + handOffsetZ };
        const particleInterval = 1;
        if (currentTick - (wallRunInfo.lastParticleTime || 0) >= particleInterval) {
            dimension.spawnParticle("parkour:wallrun_dust", handParticlePos);
            wallRunInfo.lastParticleTime = currentTick;
        }

        const footOffsetY = 0.1;
        const footOffsetDistance = 0.25;
        const footOffsetX = sideDirX * footOffsetDistance;
        const footOffsetZ = sideDirZ * footOffsetDistance;
        const footParticlePos = { x: location.x + footOffsetX, y: location.y + footOffsetY, z: location.z + footOffsetZ };
        const footParticleInterval = 2;
        if (currentTick - (wallRunInfo.lastFootParticleTime || 0) >= footParticleInterval) {
            dimension.spawnParticle("parkour:wallrun_foot_dust", footParticlePos);
            wallRunInfo.lastFootParticleTime = currentTick;
        }

        const soundInterval = 3;
        if (currentTick - wallRunInfo.lastSoundTime >= soundInterval) {
            const blockSound = BLOCK_SOUND_MAPPING[blockSide?.typeId] || DEFAULT_SOUND.sound;
            const soundConfig = SOUND_CONFIGS[blockSound] || DEFAULT_SOUND;
            dimension.playSound(blockSound, location, { pitch: soundConfig.pitch, volume: soundConfig.volume });
            wallRunInfo.lastSoundTime = currentTick;
        }
    }

    endWallRun(player, currentTick) {
        if (this.effectTimers.has(player.name)) {
            system.clearRun(this.effectTimers.get(player.name));
            this.effectTimers.delete(player.name);
        }
        if (this.fallDamageTimers.has(player.name)) {
            system.clearRun(this.fallDamageTimers.get(player.name));
            this.fallDamageTimers.delete(player.name);
        }

        player.setProperty("move:is_wallrunning", false);

        const state = this.stateManager.getState(player.name);
        if (!state) return;

        if (state.roll) {
            this.resetFallState(state.roll);
        }

        if (state.roll && !state.roll.isRolling && !state.roll.rollInputReceived) {
            const timerId = system.runTimeout(() => {
                if (!state.roll?.isRolling && !state.roll?.rollInputReceived) {
                    player.triggerEvent("move:fall_damage_true");
                }
                this.fallDamageTimers.delete(player.name);
            }, 10);
            this.fallDamageTimers.set(player.name, timerId);
        }

        this.stateManager.clearWallRunState(player.name, currentTick);
    }

    resetFallState(rollState) {
        Object.assign(rollState, {
            isInAir: false,
            validFallHeight: false,
            rollInputReceived: false,
            fallStartY: 0
        });
    }
}

export default WallRunHandler;