import { world, system } from '@minecraft/server';
import PlayerStateManager from './managers/PlayerStateManager.js';
import SlideHandler from './handlers/SlideHandler.js';
import RollHandler from './handlers/RollHandler.js';
import VaultHandler from './handlers/VaultHandler.js';
import WallRunHandler from './handlers/WallRunHandler.js';

const stateManager = PlayerStateManager;
const slideHandler = new SlideHandler(stateManager);
const rollHandler = new RollHandler(stateManager);
const vaultHandler = new VaultHandler(stateManager);
const wallRunHandler = new WallRunHandler(stateManager);

const awaitingPreWarm = new Map();

world.afterEvents.playerSpawn.subscribe((event) => {
    const player = event.player;
    if (!event.initialSpawn) return;

    try {
        awaitingPreWarm.set(player.name, {
            player,
            initialPos: { x: player.location.x, y: player.location.y, z: player.location.z },
            attempts: 0
        });
        console.log(`Awaiting movement for pre-warm: ${player.name} (initial spawn)`);
    } catch (error) {
        console.error(`Failed to setup pre-warm for ${player.name}: ${error}`);
    }
});

system.runInterval(() => {
    for (const [playerName, data] of awaitingPreWarm) {
        const { player, initialPos, attempts } = data;

        if (!player.isValid()) {
            awaitingPreWarm.delete(playerName);
            continue;
        }

        const currentPos = player.location;
        const hasMoved = (
            Math.abs(currentPos.x - initialPos.x) > 0.1 ||
            Math.abs(currentPos.y - initialPos.y) > 0.1 ||
            Math.abs(currentPos.z - initialPos.z) > 0.1
        );

        if (hasMoved) {
            try {
                rollHandler.preWarmRoll(player);
                console.log(`Pre-warmed system for ${playerName} after movement`);
                awaitingPreWarm.delete(playerName);
            } catch (error) {
                console.error(`Failed to pre-warm for ${playerName} after movement: ${error}`);
                awaitingPreWarm.delete(playerName);
            }
        } else {
            data.attempts++;
            if (data.attempts > 1000) {
                try {
                    rollHandler.preWarmRoll(player);
                    console.log(`Pre-warmed system for ${playerName} after 1000-tick timeout (no movement)`);
                    awaitingPreWarm.delete(playerName);
                } catch (error) {
                    console.error(`Failed to pre-warm for ${playerName} after timeout: ${error}`);
                    awaitingPreWarm.delete(playerName);
                }
            }
        }
    }
}, 1);

system.runInterval(() => {
    const currentTime = system.currentTick;

    for (const player of world.getPlayers()) {
        if (player.isCreative || player.isSpectator) continue;

        const state = stateManager.getState(player.name);
        if (!state) continue;

        try {
            slideHandler.handleSlide(player, currentTime);
            rollHandler.handleRoll(player, currentTime);
            vaultHandler.handleVault(player, currentTime);
            wallRunHandler.handleWallRun(player, currentTime);
        } catch (error) {
            console.error(`Error handling player ${player.name} at tick ${currentTime}: ${error}`);
        }
    }
}, 1);
